#ifndef __DEVICELIST_H__
#define __DEVICELIST_H__

/**
 * File Name  : devicelist.h
 *
 * Description: Functions to maintain a list of available HAL devices
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 *
 * Parts used from halevt, Copyright (C) 2007  Patrice Dumas <pertusus at free dot fr>
 */


//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

#include <hal/libhal.h>

G_BEGIN_DECLS


//----------------------------------------------------------------------------
// Definitions
//---------------------------------------------------------------------------- 


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------

typedef struct hal_device_property
{
  char *key;
  char **values;
  struct hal_device_property *next;
} hal_device_property;

typedef struct hal_device
{
   char *udi;
   struct hal_device_property *properties;
   struct hal_device *next;
} hal_device;


//----------------------------------------------------------------------------
// Global Constants
//----------------------------------------------------------------------------


//============================================================================
// Public Functions
//============================================================================

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_add_device
 *
 * @brief  Add udi to device list
 *
 * @param  ctx HAL context
 * @param  udi Unique Device Id
 *
 * @return 
 *
 *--------------------------------------------------------------------------*/
hal_device *hal_device_list_add_device(LibHalContext *ctx, const char *udi);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_get_property
 *
 * @brief  Return the property of the given device
 *
 * @param  property 
 * @param  device
 *
 * @return 
 *
 *--------------------------------------------------------------------------*/
hal_device_property *hal_device_list_get_property(const char *property, const hal_device *device);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_find_device
 *
 * @brief  Return the device pointed to by udi
 *
 * @param  udi Unique Device Id
 *
 * @return 
 *
 *--------------------------------------------------------------------------*/
hal_device *hal_device_list_find_device(const char *udi);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_get_root
 *
 * @brief  Get device llst root 
 *
 * @param  --
 *
 * @return Returns root of device list
 *
 *--------------------------------------------------------------------------*/
hal_device *hal_device_list_get_root();

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_remove_device
 *
 * @brief  Remove udi from device list
 *
 * @param  udi Unique Device Id
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean hal_device_list_remove_device(const char *udi);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_remove_property
 *
 * @brief  Remove property of udi from device list
 *
 * @param  udi Unique Device Id
 * @param  key
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean hal_device_list_remove_property(const char *udi, const char *key);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_device_list_set_property
 *
 * @brief  Set property of udi from device list
 *
 * @param  udi Unique Device Id
 * @param  key
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean hal_device_list_set_property(const char *udi, const char *key);

/**---------------------------------------------------------------------------
 *
 * Name :  hal_print_all_devices
 *
 * @brief  Print all available devices (for debugging)
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void hal_print_all_devices();

/**---------------------------------------------------------------------------
 *
 * Name :  hal_count_devices
 *
 * @brief  Return the number of devices available
 *
 * @return Number of devices
 *
 *--------------------------------------------------------------------------*/
int hal_count_devices();

G_END_DECLS

#endif /* __DEVICELIST_H__ */
