#ifndef __PROCESS_H__
#define __PROCESS_H__

/**
 * File Name  : process.h
 *
 * Description: Process handling functions
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 */


//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

#include <glib.h>
#include <gdk/gdk.h>
 
G_BEGIN_DECLS


//----------------------------------------------------------------------------
// Definitions
//---------------------------------------------------------------------------- 

#define PS_WAIT_STARTED     (1 << 0)
#define PS_WAIT_EXIT        (1 << 1)
#define PS_RESPAWN          (1 << 2)


//----------------------------------------------------------------------------
// Forward Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------

enum proc_state
{ 
    STATE_IDLE,
    STATE_STARTING,
    STATE_RUNNING,
    STATE_COMPLETED
};

typedef struct
{
    gchar        *command;
    gchar        *working_dir;
    gchar        *ipc_service;
    gint         flags;
    gboolean     is_multidoc;
    GPid         pid;
    GCallback    startup_callback; /* not called when respawning */
    GCallback    exit_callback;    /* not called when respawning */
    enum proc_state state;
} proc_t;


//----------------------------------------------------------------------------
// Global Constants
//----------------------------------------------------------------------------


//============================================================================
// Public Functions
//============================================================================

/**---------------------------------------------------------------------------
 *
 * Name :  process_add
 *
 * @brief  Add a process to the queue. The queue is served First In First Out.
 *
 * @param  application          Full path of the application
 * @param  working_dir          Working directory, or NULL to inherit parent's
 * @param  startup_callback     Callback function when received startupCopmlete, or NULL
 * @param  exit_callback        Callback function when processes was terminated, or NULL
 * @param  flags                Bit field PS_WAIT_STARTED | PS_WAIT_EXIT | PS_RESPAWN
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean process_add(const char* command, const char* working_dir, GCallback startup_callback, GCallback exit_callback, gint flags);

/**---------------------------------------------------------------------------
 *
 * Name :  process_start
 *
 * @brief  Start a process without queuing
 *
 * @param  application          Full path of the application
 * @param  working_dir          Working directory, or NULL to inherit parent's
 * @param  startup_callback     Callback function when received startupCopmlete, or NULL
 * @param  exit_callback        Callback function when processes was terminated, or NULL
 * @param  flags                Bit field PS_WAIT_STARTED | PS_RESPAWN
 *
 * @return Process created and started, or NULL on error.
 *
 *--------------------------------------------------------------------------*/
proc_t *process_start(const char* command, const char* working_dir, GCallback startup_callback, GCallback exit_callback, gint flags);

/**---------------------------------------------------------------------------
 *
 * Name :  process_stop
 *
 * @brief  Stop a running process
 *
 * @param  process              The process to be stopped
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void process_stop(proc_t *proc);

/**---------------------------------------------------------------------------
 *
 * Name :  process_activate
 *
 * @brief  Activate a running process  (set to top)
 *
 * @param  application Full path of the application
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean process_activate(const char* application);

/**---------------------------------------------------------------------------
 *
 * Name :  process_activate_ctb
 *
 * @brief  Activate content browser (set to top)
 *
 * @param  --
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean process_activate_ctb(void);

/**---------------------------------------------------------------------------
 *
 * Name :  process_startup_complete
 *
 * @brief  Call when application has finished starting
 *
 * @param  application Full path of the application
 * @param  pid         Process ud 
 * @param  is_multidoc TRUE if multiple documents are supported, FALSE otherwise 
 * @param  ipc_service IPC service
 * @param  window      Window id
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean process_startup_complete(const char *application, gint pid, gboolean is_multidoc, 
                                  const char *ipc_service, gint window);

proc_t *process_get_by_name(const char *application);

proc_t *process_get_by_pid(GPid pid);

#if (TESTING_ON)
void print_process_list();
#endif

G_END_DECLS

#endif /* __PROCESS_H__ */
