#ifndef __SYSTEM_H__
#define __SYSTEM_H__

/**
 * File Name  : system.h
 *
 * Description: System handling functions
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 */


//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

G_BEGIN_DECLS


//----------------------------------------------------------------------------
// Definitions
//---------------------------------------------------------------------------- 

enum state_device
{
    STATE_DEVICE_UNKNOWN = 0,
    STATE_DEVICE_STARTING,
    STATE_DEVICE_STARTED,
    STATE_DEVICE_STOPPING
};

enum state_card
{
    STATE_CARD_UNKNOWN = 0,
    STATE_CARD_INACCESSIBLE,
    STATE_CARD_EJECTED,
    STATE_CARD_UNMOUNTING,
    STATE_CARD_UNMOUNTED,
    STATE_CARD_INDEXING,
    STATE_CARD_MOUNTED
};

enum state_charge
{
    STATE_CHARGE_UNKNOWN = 0,
    STATE_CHARGE_LOW,
    STATE_CHARGE_CHARGING,
    STATE_CHARGE_DISCHARGING,
    STATE_CHARGE_FULL
};

enum state_stylus
{
    STATE_STYLUS_UNKNOWN = 0,
    STATE_STYLUS_IDLE,
    STATE_STYLUS_SCRIBBLING
};

enum state_power
{
    STATE_POWER_UNKNOWN = 0,
    STATE_POWER_OFF,
    STATE_POWER_RUN,
    STATE_POWER_IDLE,
    STATE_POWER_STANDBY
};

enum led_color
{
    COLOR_OFF = 0,  //0
    COLOR_BLUE,     //1
    COLOR_GREEN,    //2
    COLOR_CYAN,     //3
    COLOR_RED,      //4
    COLOR_MAGENTA,  //5
    COLOR_YELLOW,   //6
    COLOR_WHITE     //7   
};

enum rotate_direction
{
    ROTATE_UNKOWN = 0,
    ROTATE_CLOCKWISE,
    ROTATE_ANTICLOCKWISE
};

enum display_orientation
{
    ORIENTATION_PORTRAIT = 0,
    ORIENTATION_LANDSCAPE_CLOCKWISE,
    ORIENTATION_LANDSCAPE_ANTICLOCKWISE
};

enum busy_dialog
{
    BUSY_DIALOG_NONE = 0,
    BUSY_DIALOG_DELAYED,
    BUSY_DIALOG_DIRECT
};


//----------------------------------------------------------------------------
// Forward Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Global Constants
//----------------------------------------------------------------------------


//============================================================================
// Public Functions
//============================================================================

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_services
 *
 * @brief  Initialize the system services. Find micro driver and set up 
 *         battery level polling.
 *
 * @param  --
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void sys_set_services(void);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_get_battery
 *
 * @brief  Return battery level and charge information
 *
 * @param  [out] level    Set to current battery charge in percent
 * @param  [out] state    Charge state
 * @param  [out] timeleft Minutes to emtpy (discharging/low) or -1 (charging/full)
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_get_battery(int *level, enum state_charge *state, int *timeleft);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_get_device_capabilities
 *
 * @brief  Return device capabilities
 *
 * @param  has_stylus 
 * @param  has_wifi 
 * @param  has_bluetooth 
 * @param  has_3g
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void sys_get_device_capabilities(gboolean *has_stylus, gboolean *has_wifi, gboolean *has_bluetooth, gboolean *has_3g);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_busy
 *
 * @brief  Set the busy state for eg LED indication
 *
 * @param  is_busy TRUE for busy, FALSE otherwise
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_busy(gboolean is_busy);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_beeper_config
 *
 * @brief  Enable/disable audible feedback by applications
 *
 * @param  use_sound TRUE for beeps, FALSE to mute
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_beeper_config(gboolean use_sound);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_beeper_config
 *
 * @brief  Enable/disable audible feedback
 *  
 * @param  duration_ms Length of tone in ms (use intervals of 10ms)
 * @param  tone Beep tone "high" or "low"
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_beeper(gint duration_ms, const char *tone);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_pageturn_inverted
 *
 * @brief  Setup direction to use for the flipbar.
 *  
 * @param  is_pageturn_inverted  TRUE or FALSE 
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_pageturn_inverted(gboolean is_inverted);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_rotate_direction
 *
 * @brief  Setup direction to use for rotation. To actually rotate the screen,
 *         call sys_set_orientation().
 *  
 * @param  direction Rotate "clockwise" or "anticlockwise"
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_rotate_direction(const char *direction);
                                  
/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_orientation
 *
 * @brief  Change the orientation of the screen. The rotation direction is set
 *         by sys_set_rotate_direction().
 *  
 * @param  orientation Set to "portrait", "landscape" or "toggle"
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_orientation(const char *orientation);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_sensor_feedback
 *
 * @brief  Enable/disable audible and visual feedback by applications
 *  
 * @param  use_light TRUE to flash on sensor presses, FALSE to mute
 * @param  use_sound TRUE to beep on sensor presses, FALSE to mute
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_sensor_feedback(gboolean use_light, gboolean use_sound);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_sensor_lock_config
 *
 * @brief  Setup which sensor(s) to block when locked. To actually lock the
 *         sensors, call sys_set_sensor_lock().
 *  
 * @param  lock_left   TRUE to lock sensor, FALSE to keep unlocked
 * @param  lock_right  TRUE to lock sensor, FALSE to keep unlocked
 * @param  lock_middle TRUE to lock sensor, FALSE to keep unlocked
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_sensor_lock_config(gboolean lock_left, gboolean lock_right, gboolean lock_middle);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_sensor_lock
 *
 * @brief  Block the sensor input
 *  
 * @param  lock_mode Set to "lock", "unlock" or "toggle"
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_sensor_lock(const char *lock_mode);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_rgb_led
 *
 * @brief  Set the main RGB led
 *  
 * @param  color    Color value
 * @param  flash_ms Length of flash duration in ms (with intervals of 500ms)
 *
 * LED color values:
 *    0: off    4: red
 *    1: blue   5: magenta
 *    2: green  6: yellow
 *    3: cyan   7: white  
 * 
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_rgb_led(enum led_color color, gint flash_ms);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_set_timezone_offset
 *
 * @brief  Set the offset from UTC time 
 *
 * @param  tz_offset_sec Offset from UTC in seconds East of Greenwich
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_set_timezone_offset(gint tz_offset_sec);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_eject_card
 *
 * @brief  Unmount all volumes etc before the SD care can be safely ejected
 *
 * @param  silent [Input] TRUE to suppress dialogs, FALSE otherwise 
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void sys_eject_card(gboolean silent);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_update_battery
 *
 * @brief  Update the battery level and state
 *
 * @param  battery Battery state as read from sysfs
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void sys_update_battery(const char *battery);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_update_hardware
 *
 * @brief  Update the hardware state
 *
 * @param  hardware Hardware state as readf rom sysfs
 *
 * @return --
 *
 *--------------------------------------------------------------------------*/
void sys_update_hardware(const char *hardware);

/**---------------------------------------------------------------------------
 *
 * Name :  sys_usb_connect
 *
 * @brief  Establish USB connection by exporting file system
 *
 * @param  --
 *
 * @return TRUE for success, FALSE for failure
 *
 *--------------------------------------------------------------------------*/
gboolean sys_usb_connect(void);

void sys_set_card(enum state_card new_state);
void sys_set_idle_mode(gboolean idle_enabled);
void sys_set_standby_time(guint time_sec);
void sys_set_standby_mode(gboolean standby_if_plugged);
void sys_reset_idle_time(void);
void sys_usb_no_connect(void);
void sys_standby(void);
void sys_restart(void);
gboolean sys_request_popup(const char *state);
gboolean sys_set_power(const char *device, gint mode);
gboolean sys_get_power(const char *device, gint *mode);
guint sys_get_orientation(void);
gboolean sys_get_pageturn_inverted(void);
gfloat sys_get_display_vcom(void);
void sys_install_drz(void);
enum state_device sys_get_device_state(void);
enum state_power sys_get_power_state(void);
enum state_card sys_get_card(void);
void sys_set_device_state(enum state_device);
gboolean sys_has_stylus(void);
gboolean sys_has_network(void);
gboolean sys_is_emulator(void);
void sys_update_rgb_led(void);
gboolean sys_spawn_async(const char *command);
gboolean sys_spawn_async_with_callback(const char *command, GChildWatchFunc callback, gpointer data);
gboolean sys_spawn_sync(const char *command);
gboolean sys_is_dr800(void);

gboolean sys_get_enable_index_splash(void);
void     sys_set_enable_index_splash(gboolean value);
gboolean sys_set_index_with_metadata(gboolean value);
gint     sys_battery_level(void);
void     sys_starting_finished(void);

G_END_DECLS

#endif /* __SYSTEM_H__ */
