#ifndef __TASKS_H__
#define __TASKS_H__

/**
 * File Name  : tasks.h
 *
 * Description: Task list handling functions
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 */


//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

#include <liberipc/eripc.h>
#include <X11/Xlib.h>
#include "process.h"

G_BEGIN_DECLS


//----------------------------------------------------------------------------
// Definitions
//---------------------------------------------------------------------------- 


//----------------------------------------------------------------------------
// Forward Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Global Constants
//----------------------------------------------------------------------------


//============================================================================
// Public Functions
//============================================================================

/**---------------------------------------------------------------------------
 *
 * Name :  task_start
 *
 * @brief  Start a task
 *
 * @param  command_line Full command line to spawn, quoted as g_shell_quote
 * @param  working_dir  Working directory or NULL to inherit from sysd
 * @param  label        Text label shown under icon in popup menu
 * @param  image        Full path to icon shown in popup menu
 *
 * @return Window ID
 *
 *--------------------------------------------------------------------------*/
gboolean task_start(const char *command_line, 
                    const char *working_dir, 
                    const char *label, 
                    const char *image, 
                    eripc_context_t *context, 
                    const char *message_id);

/**---------------------------------------------------------------------------
 *
 * Name :  task_activate
 *
 * @brief  Activate a running task
 *
 * @param  command_line Full command line of task, quoted as g_shell_quote
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_activate(const char *command_line);

void task_activate_by_xid(int xid);

/**---------------------------------------------------------------------------
 *
 * Name :  task_service_by_window
 *
 * @brief  Get service of task with given window
 *
 * @param  xid          Window ID
 *
 * @return Returns IPC service
 *
 *--------------------------------------------------------------------------*/
const gchar *task_service_by_window(Window xid);


/**---------------------------------------------------------------------------
 *
 * Name :  task_stop
 *
 * @brief  Stop a running task
 *
 * @param  command_line Full command line of task, quoted as g_shell_quote
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_stop(const char *command_line,
                   eripc_context_t *context, 
                   const char *message_id);

gboolean task_rename(int xid, const char *new_document, const char* new_label);

/**---------------------------------------------------------------------------
 *
 * Name :  task_add
 *
 * @brief  Add a task which is already running
 *
 * @param  application  Basename of the application
 * @param  document     Full path of the document
 * @param  label        Text label shown under icon in popup menu
 * @param  image        Full path to icon shown in popup menu
 * @param  ipc_service  IPC service name of the application
 * @param  pid          Process ID
 * @param  xid          Window ID
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_add(const char *application, 
                  const char *document, 
                  const char *label, 
                  const char *image, 
                  const char *ipc_service, 
                  gint        pid, 
                  Window      xid);

/**---------------------------------------------------------------------------
 *
 * Name :  task_cleanup_pid
 *
 * @brief  Cleanup tasks for process
 *
 * @param  pid Process that has exited
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_cleanup_pid(GPid pid);

/**---------------------------------------------------------------------------
 *
 * Name :  task_cleanup_window
 *
 * @brief  Cleanup tasks for window
 *
 * @param  xid Window that was closed
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_cleanup_window(Window xid);

/**---------------------------------------------------------------------------
 *
 * Name :  task_startup_completed
 *
 * @brief  Call when a task has finished starting
 *
 * @param  proc Process
 * @param  xid  Window id
 *
 * @return Returns TRUE on success, FALSE on failure
 *
 *--------------------------------------------------------------------------*/
gboolean task_startup_completed(proc_t *proc, Window xid);

char* testing_task_get_list();

#if (TESTING_ON)
void print_task_list();
#endif

G_END_DECLS

#endif /* __TASKS_H__ */
