/*
 * File name: locales.c
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 */

//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

#include "config.h"

// system include files, between < >
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>

// ereader include files, between < >

// local include files, between " "
#include "log.h"
#include "locales.h"


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Global Constants
//----------------------------------------------------------------------------

#define LOCALE_NAME_FILE    "language.name"
#define LOCALE_LIB_DIR      "/usr/lib/locale"


//----------------------------------------------------------------------------
// Static Variables
//----------------------------------------------------------------------------


//============================================================================
// Local Function Definitions
//============================================================================

static gint compare_by_language_name(gconstpointer a,gconstpointer b);


//============================================================================
// Functions Implementation
//============================================================================

GSList *locales_list_create()
{
    DIR             *dirp;
    struct dirent   *dp;
    char             language[64]       = {0};
    GSList          *locales_list       = NULL;
    char            *langpath           = NULL;
    FILE            *langfile           = NULL;
    localeEntry     *locale             = NULL;
    
    LOGPRINTF("entry");
   
    // open locales directory 
    dirp = opendir(LOCALE_LIB_DIR);
    if (dirp == NULL)
    {
        ERRORPRINTF("could not open directory '%s'", LOCALE_LIB_DIR);
    }

    // scan directory and store its contents except "." and ".." 
    while ((dp = readdir(dirp)) != NULL) 
    {
        // exclude "." and ".."
        if ((dp->d_name[0] != '.' || (dp->d_name[1] != '\0' &&
            (dp->d_name[1] != '.' ||  dp->d_name[2] != '\0'))))
        {
            // reset language name
            language[0]='\0';

            // open language name file
            langpath = g_strconcat(LOCALE_LIB_DIR, "/", dp->d_name, "/", LOCALE_NAME_FILE, NULL);
            langfile = fopen(langpath, "r");
            if (langfile != NULL)
            {
                // read language name from file
                fgets(language, 63, langfile);
                fclose(langfile);
            }
            g_free(langpath);

            // add locale to list
            locale = g_new0 (localeEntry, 1);
            locale->code = g_strdup(dp->d_name);
            if (strlen(language) > 0)
            {
                // found language name, remove trailing linefeed
                if (language[strlen(language)-1] == '\n') 
                {
                    language[strlen(language)-1]='\0';
                }
                locale->language = g_strdup(language);
            }
            else 
            {
                // no native name found, store locale code
                locale->language = g_strdup(dp->d_name);
            }
                
            locales_list = g_slist_prepend(locales_list, locale);
        }
    }
    closedir(dirp);
    
    // sort locales by language name
    locales_list = g_slist_sort(locales_list, compare_by_language_name);

    return locales_list;
}



GSList *locales_list_codes(GSList *locales_list)
{
    GSList *code_list = NULL;
    
    LOGPRINTF("entry");

    while (locales_list != NULL)
    {
        localeEntry *locale = locales_list->data;
        code_list = g_slist_append(code_list, locale->code);
        locales_list = locales_list->next;
    }

    return code_list;
}


GSList *locales_list_languages(GSList *locales_list)
{
    GSList *lang_list = NULL;
    
    LOGPRINTF("entry");

    while (locales_list != NULL)
    {
        localeEntry *locale = locales_list->data;
        lang_list = g_slist_append(lang_list, locale->language);
        locales_list = locales_list->next;
    }

    return lang_list;
}


void locales_list_print(GSList *locales_list)
{
    LOGPRINTF("entry");

    while (locales_list != NULL)
    {
#if (ERROR_ON)
        localeEntry *locale = locales_list->data;
        
        ERRORPRINTF("code: %s, language: %s", 
                  (char*) locale->code,
                  (char*) locale->language);
#endif        
        
        locales_list = locales_list->next;
    }
}


void locales_list_free(GSList *locales_list)
{
    LOGPRINTF("entry");

    g_return_if_fail(locales_list != NULL);

    while (locales_list != NULL)
    {
      localeEntry *locale = locales_list->data;
        
      // free
      g_free(locale->code);
      g_free(locale->language);
      g_free(locale);
            
      locales_list =locales_list->next;
    }
    
    g_slist_free(locales_list);
    locales_list = NULL;
}


static gint compare_by_language_name(gconstpointer a,gconstpointer b)
{
    localeEntry *aa = (localeEntry *)a;
    localeEntry *bb = (localeEntry *)b;
//    LOGPRINTF("sort: %s, %s", (gchar *) (aa->language),(gchar *) (bb->language));
    return strcmp((gchar *) aa->language,(gchar *) bb->language);
}
