
#include "config.h" 

// system include files, between < >
#include <glib.h>
#include <hal/libhal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/statvfs.h>

// ereader include files, between < >
#include <liberipc/eripc.h>
#include <liberutils/display_utils.h>
#include <liberkeyb/erkeyb-client.h>

// local include files, between " "
#include "log.h"
#include "connections.h"
#include "conf.h"
#include "display.h"
#include "hal.h"
#include "ipc.h"
#include "wacom.h"
#include "xwindow.h"
#include "tasks.h"

#define write_ionkbd(file, ...)     write_sysfs(g_serio,  file, __VA_ARGS__)
#define ADOBE_SYNC                  "/usr/bin/adobe-sync.sh"

extern void sync(void);

gchar* g_serio;
eripc_client_context_t *eicc;


//============================================================================
// Local Function Definition
//============================================================================

gboolean      sys_spawn_sync             (const char *command);
static gboolean      write_sysfs                (const char *device, const char *filename,  const char *buffer);
static gchar         *find_dir                  (const char *dir, const char *file, gboolean recurse);
static void          on_poweroff                (void);
static void          on_poweroff_continue       (gpointer data);
static void          report_unmounting_volumes  (void);
static void          do_unmount                 (gpointer callback_function, gpointer user_data);


int main() {
    LOGPRINTF("before init");

    ipc_set_services(); //todo: ipc.c

    // init bus client
    eicc = eripc_client_context_new("irexshutdown",
                                    "1.0",
                                    "com.irexnet.irexshutdown",
                                    "/com/irexnet/irexshutdown",
                                    "com.irexnet.irexshutdown",
                                    NULL);
    erkeyb_client_init();    

    LOGPRINTF("before send signal");
    // let application prepare for standby
    eripc_send_signal_varargs(eicc->context, 
                              ERIPC_BUS_SESSION,
                              DBUS_PATH,
                              DBUS_INTERFACE,
                              "sysPrepareStandby",
                              ERIPC_TYPE_INVALID);


    LOGPRINTF("before conn_stop");
    // stop connections
    conn_stop();   // todo:connections.c

    // set a nice splash screen
    LOGPRINTF("before set splash");
    eripc_send_string(eicc->context,
	                       NULL,
	                       NULL,
	                       ERIPC_BUS_SESSION,
	                       DBUS_SERVICE_MENU,
	                       "setSplashShow", "shutdown");

    
    // 3 sec = PREPARE_STANDBY_TIMEOUT_MS  2500 + CHECK_STANDBY_TIMEOUT_MS    500, see system.c
    sleep(3); 

    // proceed with poweroff
    on_poweroff(); 

    return(0);
}


static void on_poweroff() {
    LOGPRINTF("on_poweroff");

    // export gconf to INI here so changes made during "prepare standby" as saved
    sys_spawn_sync(ERCONFTOOL_EXPORT);

    // commit buffer cache to disk
    sync();
    
    // do any connection related suspend work
    LOGPRINTF("conn standby");
    conn_on_enter_standby();  // todo: connections.c

    /* now ask for an unmount */
    LOGPRINTF("unmount");

    eripc_sysd_unmount_sd_card(eicc);
    do_unmount(on_poweroff_continue, NULL);
}

static void on_poweroff_continue(gpointer data) {
    LOGPRINTF("on_poweroff_continue");

    // enable/unlock all three sensors before power off
    find_dir("/sys/devices", "serio", FALSE); //todo g_serio=...
    write_ionkbd("sensor_enable", "7"); 

    // power off and wait
    LOGPRINTF("yahoo!!");
    sys_spawn_sync("shutdown -h now");

    while (1)
    {
        g_usleep(G_USEC_PER_SEC * 1);
    }
}


static void do_unmount(gpointer callback_function, gpointer user_data)
{
    LOGPRINTF("do_unmount");

    // have applications close their files
    report_unmounting_volumes();

    // run export settings and wait
    sys_spawn_sync(ERCONFTOOL_EXPORT);

    // synchronize with Adobe Digital Editions
    sys_spawn_sync(ADOBE_SYNC);

    // unmount volumes
    hal_unmount_all_volumes(callback_function, user_data);

    LOGPRINTF("unmount done");
}

static void report_unmounting_volumes()
{
    hal_device *device = hal_device_list_get_root();
    hal_device_property *property;

    while (device != NULL)
    {
        property = hal_device_list_get_property("volume.mount_point", device);
        if ((property != NULL) && (*property->values != NULL) && ((*property->values)[0] != '\0'))
        {
//            LOGPRINTF("about to unmount: %s", *property->values);

	    eripc_send_signal_varargs(eicc->context, 
                              ERIPC_BUS_SESSION,
                              DBUS_PATH,
                              DBUS_INTERFACE,
                              "sysPrepareUnmount",
                              ERIPC_TYPE_STRING, *property->values,
                              ERIPC_TYPE_INVALID);
            
        }
        device = device->next;
    }
}

//// helper methods

static char *find_dir(const char *dir, const char *name, gboolean recurse)
{
    struct  dirent **namelist;
    char   *full_name = NULL;
    int     i         = 0;
    int     count     = 0;

    count = scandir(dir, &namelist, 0, alphasort);
    for(i = 0; i < count; i++) 
    {
        char *filename = namelist[i]->d_name;
        
        if (filename && (strcmp(filename, ".") != 0) && (strcmp(filename, "..") != 0)) 
        {
            char fullpath[512] = "";
            strcat(fullpath, dir);
            strcat(fullpath, "/");
            strcat(fullpath, filename);

            struct stat attributs;
            if (lstat(fullpath, &attributs) == -1)
            {
//                WARNPRINTF("Error reading attributes of '%s'", fullpath);
            }

            if S_ISDIR(attributs.st_mode)
            {
                if (name != NULL)
                {
                    if (strncmp(filename, name, strlen(name)) == 0)
                    {
                        // dir found
                        full_name = g_strdup(fullpath);
                        break;
                    }
                    else if (recurse && name != NULL) 
                    {
                        // recursive search 
                        full_name = find_dir(fullpath, name, recurse);
                        if (full_name)
                        {
                            // dir found
                            break;
                        }
                        g_free(full_name);
                    }
                }
            } 
        }
    }
    
    g_free(namelist);
    return full_name;
}

static gboolean write_sysfs(const char *device, const char *filename, const char *buffer)
{
//    LOGPRINTF("entry");
    
    gboolean retval = FALSE;
    char *file = NULL;
    gint fd = 0;
    gint result = 0;
    
    if (!device)
    {
        return FALSE;
    }
    
    file = g_strconcat(device, "/", filename, NULL);
    
    fd = open(file, O_WRONLY);
    if (fd >= 0)
    {
        result = write(fd, buffer, strlen(buffer));
        close(fd);
        
        if (result > -1)
        {
            retval = TRUE;
        }
        else
        {
//            WARNPRINTF("Could not write to `%s` (%d), %s", file, errno, strerror(errno));
        }
    }
    else
    {
//        WARNPRINTF("Could not open %s", filename);
    }
    
    g_free(file);
    
    return retval;
}

/* // todo:  enable this == get rid of system.c
gboolean sys_spawn_sync(const char *command)
{
    // LOGPRINTF("entry");
    
    g_return_val_if_fail(command!=NULL, FALSE);

    char *argv[4];
    gboolean retval = FALSE;
    GError *error = NULL;
    
    argv[0] = "/bin/sh";
    argv[1] = "-c";
    argv[2] = (char*) command;
    argv[3] = NULL;

    // LOGPRINTF("Spawning: %s", argv[2]);

    retval = g_spawn_sync(NULL, argv, NULL, 0, NULL, NULL, NULL, NULL, NULL, &error);
    
    if (error)
    {
        // WARNPRINTF("Execution of '%s' failed with error: %s", command, error->message);
        g_error_free(error);
        return FALSE;
    }

    return retval;
}
*/


/// connections.c

