/*
 * File Name: wacom.c
 */

/*
 * This file is part of sysd.
 *
 * sysd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * sysd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Copyright (C) 2008 iRex Technologies B.V.
 * All rights reserved.
 */

//----------------------------------------------------------------------------
// Include Files
//----------------------------------------------------------------------------

// system include files, between < >
#include <glib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>

// local include files, between " "
#include "log.h"
#include "system.h"
#include "wacom.h"


//----------------------------------------------------------------------------
// Type Declarations
//----------------------------------------------------------------------------

typedef struct {
    char *uname;
    char *cmd;
    char *help;
} command;


//----------------------------------------------------------------------------
// Constants
//----------------------------------------------------------------------------

#define NUM_COMMANDS  8
#define DEV_WACOM     "/dev/ttymxc1"


//----------------------------------------------------------------------------
// Static Variables
//----------------------------------------------------------------------------

static command commands[] = {
    {"stop",        "0", "Stop sampling"},
    {"sample-133",  "1", "Start sampling at 133 pts/sec"},
    {"sample-80",   "2", "Start sampling at 80 pts/sec"},
    {"sample-40",   "3", "Start sampling at 40 pts/sec"},
    {"survey-on",   "+", "Enable survey scan"},
    {"survey-off",  "-", "Disable survey scan"},
    {"tiltcor-on",  "t", "Enable tilt correction"},
    {"tiltcor-off", "u", "Disable tilt correction"},
};

static gboolean g_wacom_enabled = FALSE;


//============================================================================
// Local Function Definitions
//============================================================================

static gboolean send_command(char *name);
static gint port_open(char *name);
static void set_defaults(void);


//============================================================================
// Functions Implementation
//============================================================================

void wacom_scan_high()
{
    LOGPRINTF("entry");
    if (g_wacom_enabled)
    {
        wacom_resume();
        send_command("sample-133");
    }
    else
    {
        WARNPRINTF("wacom is disabled, mode not changed");
    }
}


void wacom_scan_normal()
{
    LOGPRINTF("entry");
    if (g_wacom_enabled)
    {
        wacom_resume();
        send_command("sample-80");
    }
    else
    {
        WARNPRINTF("wacom is disabled, mode not changed");
    }
}


void wacom_suspend()
{
    LOGPRINTF("entry");
    if (g_wacom_enabled)
    {
        // since the wacom switches to normal mode (8pps) automatically, don't do anything
    }
    else
    {
        WARNPRINTF("wacom is disabled, mode not changed");
    }
}


void wacom_resume()
{
    LOGPRINTF("entry");
    if (g_wacom_enabled)
    {
        // since the wacom switches to writing mode (133pps) automatically, don't do anything
    }
    else
    {
        WARNPRINTF("wacom is disabled, mode not changed");
    }
}


void wacom_tilt_correction_enable(gboolean tilt_enable)
{
    LOGPRINTF("entry");
    if (g_wacom_enabled && tilt_enable)
    {
        send_command("tiltcor-on");
    }
    else if (g_wacom_enabled && !tilt_enable)
    {
        send_command("tiltcor-off");
    }
    else
    {
        WARNPRINTF("wacom is disabled, tilt correction not changed");
    }
}


void wacom_enable()
{
    LOGPRINTF("entry");
    
    g_wacom_enabled = TRUE;
    sys_set_power("wacom", 1);
    set_defaults();
}


void wacom_disable()
{
    LOGPRINTF("entry");
    
    sys_set_power("wacom", 0);
    g_wacom_enabled = FALSE;
}


void wacom_reset()
{
    LOGPRINTF("entry");
    
    g_wacom_enabled = TRUE;
    sys_set_power("wacom_reset", 1);
    set_defaults();
}


gboolean wacom_is_enabled()
{
    return g_wacom_enabled;
}


//============================================================================
// Local Function Implementation
//============================================================================

static void set_defaults(void)
{
    wacom_scan_high();
    wacom_tilt_correction_enable(FALSE);
}


static gboolean send_command(char *name)
{
    LOGPRINTF("entry");

    g_return_val_if_fail(name != NULL, FALSE);

    gboolean retval = FALSE;
    gint fd = 0;
    gint num = 0;

    fd = port_open(DEV_WACOM);
    if (fd > 0)
    {
        gint i;
        for (i=0; i<NUM_COMMANDS; i++)
        {
            if (strcmp(name, commands[i].uname)==0)
            {
                LOGPRINTF("sending command %s, %s", commands[i].uname, commands[i].help);

                num = write(fd,commands[i].cmd, strlen(commands[i].cmd));
                LOGPRINTF("- bytes sent %d (%s)", num, commands[i].cmd);
                if (num < 0)
                {
                    ERRORPRINTF("- error sending: %d (%s)", errno, strerror(errno));
                }
                    
                retval = TRUE;
            }
        }
        close(fd);
    }
    
    return retval;
}


static gint port_open(char *name)
{
    LOGPRINTF("entry");
    
    g_return_val_if_fail(name != NULL, FALSE);
    
    gint fd;
    struct termios t;
    
    if ((fd = open(name, O_RDWR | O_NOCTTY)) < 0)
    {
        ERRORPRINTF("Can't open device %s\n", name);
        return FALSE;
    }

    tcgetattr(fd, &t);

    t.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
    t.c_oflag &= ~OPOST;
    t.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
    t.c_cflag &= ~(CSIZE|PARENB);
    t.c_cflag |= CS8|CLOCAL;
    t.c_cflag &= ~(CSTOPB);     // 1 stop bit 
    t.c_cflag &= ~(CSIZE);      // 8 data bits 
    t.c_cflag |= CS8;
    t.c_cflag &= ~(PARENB);     // no parity 
    t.c_iflag |= IXOFF;         // flow control XOff 
    t.c_cc[VMIN] = 0;           // vmin value 
    t.c_cc[VTIME] = 10;         // vtime value 

    cfsetispeed(&t, B19200);
    cfsetospeed(&t, B19200);
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd, TCSANOW, &t);
    
    return fd;
}

    
